local MAJOR = "StarLibLCDText-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LCDText = LibStub:NewLibrary(MAJOR, MINOR)
if not LCDText then return end
local core = LibStub("StarLibCore-1.0")
assert(core, MAJOR .. " requires StarLibCore-1.0")
local LibError = LibStub("StarLibError-1.0")
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibBuffer = LibStub("StarLibBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires StarLibBuffer-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LCDText.__index then
	LCDText.__index = LCDText
end

-- @name StarLCDText:New
-- @usage StarLCDText:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor A StarLibCore object
-- @param rows The number of rows
-- @param cols The number of cols
-- @param yres The y resolution for characters
-- @param xres The x resolution for characters
-- @param layers The number of layers
-- @param errorLevel Error verbosity level
-- @return A new StarLCDText object

function LCDText:New(visitor, rows, cols, yres, xres, layers, errorLevel, realBlit, data)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.error = LibError:New(MAJOR, errorLevel) 
	obj.visitor = visitor
	obj.LROWS = rows
	obj.LCOLS = cols
	obj.DROWS = rows
	obj.DCOLS = cols
	obj.LAYERS = layers
	obj.errorLevel = errorLevel
	obj.RealBlit = realBlit
	obj.data = data
	obj.transition_tick = 0

	local n = rows * cols
	
	obj.LayoutFB = {}
	obj.TransitionFB = {}
	for i = 0, layers do
		obj.LayoutFB[i] = LibBuffer:New(MAJOR .. ".LayoutFB", n, self.errorLEvel)
		obj.TransitionFB[i] = LibBuffer:New(MAJOR .. ".TransitionFB", n, self.errorLevel)
	end
	obj.DisplayFB = LibBuffer:New(MAJOR .. ".DisplayFB", n, self.errorLevel)
	
	return obj	
	
end

-- @name StarLCDText:Del
-- @usage StarLCDText:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LCDText:Del(lcd)
	if not lcd then lcd = self end
	
	if type(lcd) == "table" then
		pool[lcd] = true
		for i, v in ipairs(objects) do
			if v == lcd then
				objectsDict[lcd] = nil
				tremove(objects, i)
			end
		end
		lcd.error:Del()
	end
end

function LCDText:Clear()
	for i = 0, self.LROWS * self.LCOLS - 1 do
		for l = 0, self.LAYERS - 1 do
			self.LayoutFB[l].buffer[i] = ' '
			self.TransitionFB[l].buffer[i] = ' '
		end
	end
	--self:TextBlit(0, 0, self.LROWS, self.LCOLS)
end

function LCDText:TextBlit(row, col, height, width)
	--local lr, lc
	--local dr, dc
	--local p1, p2
	self._fb = LibBuffer:New(MAJOR .. "._fb", self.LROWS * self.LCOLS)
	
	local r = row
	
	while r < self.LROWS and r < row + height do
		local c = col
		while c < self.LCOLS and c < col + width do
			local n = r * self.LCOLS + c
			local l = self.LAYERS - 1
			while l >= 0 do
				if self.LayoutFB[l].buffer[n] ~= ' ' then
					self._fb:Replace(n, self.LayoutFB[l].buffer[n])
					
				end
				l = l - 1
			end
			c = c + 1
		end
		r = r + 1
	end
	
	--self._fb:Copy(self.LayoutFB[0])

	--/* loop over layout rows */
	self._lr = row
	
	while self._lr < self.LROWS and self._lr < row + height do
		--/* transform layout to display row */
		self._dr = self._lr;
		--/* sanity check */
		if self._dr >= 0 and self._dr < self.DROWS then 		
			--/* loop over layout cols */
			self._lc = col
			while self._lc < self.LCOLS and self._lc < col + width do
				--/* transform layout to display column */
				self._dc = self._lc;
				--/* sanity check */
				if self._dc >= 0 and self._dc < self.DCOLS then
					--/* find start of difference */
					if self.DisplayFB.buffer[self._dr * self.DCOLS + self._dc] ~= self._fb.buffer[self._lr * self.LCOLS + self._lc] then
					--if string.sub(self.DisplayFB, dr * self.DCOLS + dc, 1) ~= string.sub(fb, lr * self.LCOLS + lc) then
						--/* find end of difference */
						self._p1 = self._dc
						self._p2 = self._p1
						while self._lc < self.LCOLS and self._lc < col + width do
							--/* transform layout to display column */
							self._dc = self._lc;
							--/* sanity check */
							if self._dc >= 0 and self._dc < self.DCOLS then
								if self.DisplayFB.buffer[self._dr * self.DCOLS + self._dc] ~= self._fb.buffer[self._lr * self.LCOLS + self._lc] then
									self._p2 = self._dc;
								end
							end
							self._lc  = self._lc + 1
						end

						--/* send to display */
						self._from = self._fb:Sub(self._lr * self.LCOLS + self._p1)
						self.DisplayFB:Merge(self._from, self._dr * self.DCOLS + self._p1, self._p2 - self._p1 + 1)
						if (self.RealBlit) then
							self._tmp = self.DisplayFB:Sub(self._dr * self.DCOLS + self._p1)
							self:RealBlit(self.data, self._dr, self._p1, self._tmp.buffer, self._p2 - self._p1 + 1)
							self._tmp:Del()
							self._tmp = nil
						end
						self._from:Del()
					end
				end
				self._lc = self._lc + 1
			end
		end
		self._lr = self._lr + 1
	end
	self._fb:Del()
end

function LCDText.TextDraw(widget, data)
	widget._lcd = widget.visitor.lcd
	widget._n  = widget.widget.row * widget._lcd.LCOLS + widget.widget.col
	widget._len = strlen(widget.buffer)
	widget._num = 1
	local layer = widget.widget.layer
    local fb;
    if widget._lcd.transitioning and widget.widget.layout_base == widget._lcd.transitionLayout then
        fb = widget._lcd.TransitionFB[layer];
    else
        fb = widget._lcd.LayoutFB[layer];
	end
	
	for i = widget._n, widget._n + widget._len do
		if i - widget.widget.row * widget._lcd.LCOLS >= widget._lcd.LCOLS then
			break
		end
		fb:Replace(i, widget.buffer:sub(widget._num, widget._num))
		widget._num = widget._num + 1
	end

	if not widget._lcd.transitioning then
		widget._lcd:TextBlit(widget.widget.row, widget.widget.col, 1, widget._len)
	end
end

function LCDText.BarDraw(widget)
	local lcdText = widget.widget.visitor.lcd
    local length = widget.length
    local row = widget.widget.row
    local col = widget.widget.col
    local layer = widget.widget.layer

    local val = math.floor((widget.val1 * length) + 0.5 );

    if(val < 0) then val = 0 end

    local n = row * lcdText.LCOLS + col;

    local fb;
    if lcdText.transitioning and widget.widget.layout_base == lcdText.transitionLayout then
        fb = lcdText.TransitionFB[layer];
    else
        fb = lcdText.LayoutFB[layer];
	end

    for x = n, n + length do
        fb.buffer[x] = ' ';
	end

    if( widget.direction == widget.DIR_EAST ) then
		for i = n, n + val - 1 do
			if mod(i, lcdText.LCOLS) >= lcdText.LCOLS then
				break
			end
            fb.buffer[i] = "#"
        end
    elseif( widget.direction == widget.DIR_WEST ) then
        for i = n, n - val + 1, -1 do
			if i / lcdText.LROWS < row then
				break
			end
			fb.buffer[i] = "#"
        end
	end
    if not lcdText.transitioning then
        lcdText:TextBlit(row, col, 1, length)
	end
end

function LCDText.HistogramDraw(widget)
	local lcdText = widget.widget.visitor.lcd

    local row = widget.widget.row
    local col = widget.widget.col
    local width = widget.width
	local height = widget.height
	local layer = widget.widget.layer

    local fb;
    if(lcdText.transitioning and widget.widget.layout_base == lcdText.transitionLayout) then
        fb = lcdText.TransitionFB[layer]
    else
        fb = lcdText.LayoutFB[layer]
	end
	
	if not fb then 
		widget.error:Print("Bad layer " .. layer)
		return 
	end

	for c = col, col + width - 1 do
		for r = row, row + height - 1 do
			fb.buffer[r * lcdText.LCOLS + c] = " "
		end
	end
	
	for c = col, col + width - 1 do
		local val
		if c - col >= #widget.history then
			val = 0
		else
			val = floor(widget.history[c - col + 1] * height + 0.5)
		end
		for r = row + height, row + height - val, -1 do
			fb.buffer[(row + r) * lcdText.LCOLS + c] = "#"
		end
	end
	
    if(not lcdText.transitioning) then
        lcdText:TextBlit(row, col, height, width);
	end
end

function LCDText:SignalTransitionStart(layout)
	self.transitioning = true; 
	self.transitionLayout = layout
end

function LCDText:SignalTransitionEnd()
	self.transitioning = false
end

function LCDText:Transition()
	if self.visitor.direction == self.visitor.TRANSITION_LEFT or
	   self.visitor.direction == self.visitor.TRANSITION_RIGHT or
	   self.visitor.direction == self.visitor.TRANSITION_BOTH then
		self:TransitionLeftRight()
	elseif self.visitor.direction == self.visitor.TRANSITION_UP or
	       self.visitor.direction == self.visitor.TRANSITION_DOWN then
		self:TransitionUpDown()
	else
		self:TransitionLeftRight()
	end
end

function LCDText:TransitionLeftRight()
	local direction = self.visitor.direction
	local col
	local left, right
	local layout = LibBuffer:New("tmplayout", self.LROWS * self.LCOLS, self.errorLevel)
	local transition = LibBuffer:New("tmptransition", self.LROWS * self.LCOLS, self.errorLevel)

	for l = self.LAYERS - 1, 0, -1 do
		for n = 0, self.LROWS * self.LCOLS - 1 do
			if self.LayoutFB[l].buffer[n] ~= ' ' then
				layout.buffer[n] = self.LayoutFB[l].buffer[n]
			end
			if self.TransitionFB[l].buffer[n] ~= ' ' then
				transition.buffer[n] = self.TransitionFB[l].buffer[n]
			end
		end
	end

    for row = 0, self.LROWS - 1 do
		local n = row * self.LCOLS
		
		
        if( direction == self.visitor.TRANSITION_LEFT or
            (direction == self.visitor.TRANSITION_BOTH and mod(row, 2) == 0)) then
            col = self.LCOLS - self.transition_tick - 1;
            left = layout;
            right = transition;
        elseif( direction == self.visitor.TRANSITION_RIGHT or direction == self.visitor.TRANSITION_BOTH) then
            col = self.transition_tick;
            left = transition;
            right = layout;
        else 
            col = self.transition_tick;
            left = transition;
            right = layout;
            self.error:Print("LCDText: Bad transition direction: " .. direction);
        end
		--self.DisplayFB:Memset(n, ' ', self.LCOLS)
		self.DisplayFB:Memcpy(n, left, n + (self.LCOLS - col), col)
		self.DisplayFB:Memcpy(n + col + 1, right, n, self.LCOLS - col - 1)
        --memset(DisplayFB + n, ' ', LCOLS);
        --memcpy(DisplayFB + n, left + n + (LCOLS - col), col);
        --memcpy(DisplayFB + n + col + 1, right + n, LCOLS - col - 1);
        if(direction ~= self.visitor.TRANSITION_BOTH) then
            self.DisplayFB.buffer[n + col] = ':';
		end
        if(self.RealBlit) then
			local fb = self.DisplayFB:Sub(n)
            self:RealBlit(self.data, row, 0, fb.buffer, self.LCOLS);
			fb:Del()
		end
		
    end

	self.transition_tick = self.transition_tick + 1
    if( self.transition_tick >= self.LCOLS ) then
        self.transitioning = false;
        self.transition_tick = 0;
        --emit static_cast<LCDEvents *>(
        --    visitor_->GetWrapper())->_TransitionFinished();
		self.visitor:TransitionFinished()
        for l = 0, self.LAYERS - 1 do
			self.LayoutFB[l]:Copy(self.TransitionFB[l])
			--self.LayoutFB[l]:Memcpy(0, self.TransitionFB[l], 0, self.LCOLS * self.LROWS)
            --memcpy(LayoutFB[l], TransitionFB[l], LCOLS * LROWS);
			self.TransitionFB[l]:Clear()
			--self.TransitionFB[l]:Memset(0, ' ', self.LCOLS * self.LROWS)
            --memset(TransitionFB[l], ' ', LCOLS * LROWS);
        end
        self:TextBlit(0, 0, self.LROWS, self.LCOLS);
    end
	
	layout:Del()
	transition:Del()
end

function LCDText:TransitionUpDown()
	do return end
    local direction = self.visitor.direction
    local row;
    local top, bottom;
    local layout = LibBuffer:New("tmplayout", LROWS * LCOLS, self.errorLevel)
	local transition = LibBuffer:New("tmptransition", LROWS * LCOLS, self.errorLevel)
 
    for l = self.LAYERS - 1, 0, -1 do
        for n = 0, self.LROWS * self.LCOLS - 1 do
            if(self.LayoutFB[l].buffer[n] ~= ' ') then
                layout.buffer[n] = self.LayoutFB[l].buffer[n];
			end
            if(self.TransitionFB[l].buffer[n] ~= ' ') then
                transition.buffer[n] = self.TransitionFB[l].buffer[n];
			end
        end
    end

    if(direction == self.visitor.TRANSITION_UP) then
        top = layout;
        bottom = transition;
        row = self.LROWS - self.transition_tick
    else 
        top = transition;
        bottom = layout;
        row = self.transition_tick
    end

    for r = 0, self.LROWS - 1 do
        local n = r * LCOLS;
        for l = 0, self.LAYERS do
			self.DisplayFB:Memset(n, ' ', self.LCOLS)
            if(r < row) then
				self.DisplayFB:Memcpy(n, top, n, self.LCOLS)
            else
				self.DisplayFB:Memcpy(n, bottom, n, self.LCOLS)
            end
        end
        if(self.RealBlit) then
            RealBlit(self.visitor.lcd, r, 0, self.DisplayFB + n, self.LCOLS);
		end
    end

	self.transition_tick = self.transition_tick + 1
    if( self.transition_tick >= self.LROWS ) then
        self.transitioning = false;
        self.transition_tick = 0;
        --emit static_cast<LCDEvents *>(
        --    visitor_->GetWrapper())->_TransitionFinished();
		self.visitor:TransitionFinished()
        for l = 0, self.LAYERS - 1 do
			self.LayoutFB[l]:Memcpy(0, self.TransitionFB[l], 0, self.LCOLS * self.LROWS)
            --memcpy(LayoutFB[l], TransitionFB[l], LCOLS * LROWS);
			self.TransitionFB[l]:Memset(0, ' ', self.LCOLS * self.LROWS)
            --memset(TransitionFB[l], ' ', LCOLS * LROWS);
        end
        self:TextBlit(0, 0, self.LROWS, self.LCOLS);
    end
	
	layout:Del()
	transition:Del()
end
